import typing as t
from datetime import datetime

from pipeline.cloud.compute_requirements import Accelerator
from pipeline.cloud.schemas import BaseModel
from pipeline.cloud.schemas.runs import RunIOType


class IOVariable(BaseModel):
    run_io_type: RunIOType

    title: str | None
    description: str | None
    examples: list[t.Any] | None
    gt: int | None
    ge: int | None
    lt: int | None
    le: int | None
    multiple_of: int | None
    allow_inf_nan: bool | None
    max_digits: int | None
    decimal_places: int | None
    min_length: int | None
    max_length: int | None
    choices: list[t.Any] | None
    dict_schema: t.List[dict] | None
    default: t.Any | None


class PipelineGet(BaseModel):
    id: str
    created_at: datetime
    updated_at: datetime

    name: str
    path: str

    minimum_cache_number: t.Optional[int]
    gpu_memory_min: t.Optional[int]
    environment_id: str

    accelerators: t.Optional[t.List[Accelerator]]

    input_variables: t.List[IOVariable]
    output_variables: t.List[IOVariable]

    _metadata: t.Optional[dict]


class PipelinePatch(BaseModel):
    minimum_cache_number: t.Optional[int]
    gpu_memory_min: t.Optional[int]
    accelerators: t.Optional[t.List[Accelerator]]



import json
import urllib.parse
from argparse import ArgumentParser, Namespace, _SubParsersAction

from websockets.exceptions import ConnectionClosedError
from websockets.sync.client import connect

from pipeline import current_configuration
from pipeline.util.logging import _print, _print_remote_log


def run_logs_parser(command_parser: "_SubParsersAction[ArgumentParser]") -> None:
    run_logs_parser = command_parser.add_parser("run", help="Get logs for a run.")
    run_logs_parser.set_defaults(func=_run_logs)

    run_logs_parser.add_argument(
        "-f",
        "--follow",
        action="store_true",
    )

    run_logs_parser.add_argument(
        "run_id",
        help="Run ID to get logs for.",
        type=str,
    )


def _run_logs(args: Namespace) -> None:
    follow = getattr(args, "follow", False)
    run_id = getattr(args, "run_id")

    url_info = urllib.parse.urlparse(current_configuration.active_remote.url)
    target_host = url_info.hostname
    target_port = url_info.port

    connection_string = urllib.parse.urljoin(
        "ws://" + target_host + ":" + str(target_port),
        f"/v3/logs/run/{run_id}?follow={follow}",
    )

    try:
        with connect(
            connection_string,
            additional_headers={
                "Authorization": f"Bearer {current_configuration.active_remote.token}"
            },
        ) as websocket:
            message = websocket.recv()
            _print(f"Showing logs for run {run_id}")
            while True:
                stream_data: dict = json.loads(json.loads(message))
                streams = stream_data.get("streams", None)
                if streams is None:
                    raise Exception("No streams in message")

                if len(streams) == 0:
                    raise Exception("No streams in message")

                for stream in streams:
                    values = stream.get("values", None)
                    if values is None:
                        raise Exception("No values in stream")

                    for value in values:
                        _print_remote_log(
                            value,
                        )
                if not follow:
                    break
                message = websocket.recv()
    except ConnectionClosedError as e:
        if e.code == 4000:
            _print(f"Run {run_id} not found", level="ERROR")



import os
from datetime import datetime

import humps
from pydantic import BaseModel as _BaseModel
from pydantic import Extra
from pydantic.generics import GenericModel as _GenericModel

CAMEL_CASE_ALIASES = bool(os.environ.get("CAMEL_CASE_ALIASES", False))


def _generate_alias(s):
    return humps.camelize(s) if CAMEL_CASE_ALIASES else s


class BaseConfig:
    alias_generator = _generate_alias
    allow_population_by_field_name = True
    # Encode datetime objects to JSON as integer timestamps
    json_encoders = {datetime: datetime.timestamp}
    orm_mode = True


class BaseModel(_BaseModel):
    class Config(BaseConfig):
        orm_mode = True
        use_enum_values = True


class Patchable(BaseModel):
    class Config:
        extra = Extra.forbid


class GenericModel(_GenericModel):
    """Base model for schemas with generic typing."""

    class Config(BaseConfig):
        pass



import importlib
import json
import math
import os
import platform
import time
import typing as t
import uuid
from importlib.metadata import version
from multiprocessing import Pool
from pathlib import Path
from tempfile import SpooledTemporaryFile
from threading import Thread
from urllib.parse import urlparse
from zipfile import ZipFile

import cloudpickle as cp
import httpx
from httpx import HTTPStatusError
from pydantic import ValidationError

from pipeline.cloud import http
from pipeline.cloud.compute_requirements import Accelerator
from pipeline.cloud.logs import tail_run_logs
from pipeline.cloud.schemas import pipelines as pipeline_schemas
from pipeline.cloud.schemas.runs import (
    Run,
    RunCreate,
    RunError,
    RunInput,
    RunIOType,
    RunOutput,
    RunState,
)
from pipeline.configuration import current_configuration
from pipeline.objects import Directory, File, FileURL, Graph
from pipeline.util.logging import _print, _print_remote_log


class PipelineRunError(Exception):
    """Error raised when there was an exception raised during a pipeline run"""

    def __init__(
        self,
        exception: str,
        traceback: t.Optional[str],
    ) -> None:
        self.exception = exception
        self.traceback = traceback
        super().__init__(self.exception, self.traceback)

    def __str__(self):
        error_msg = self.exception
        if self.traceback:
            error_msg += f"\n\nFull traceback from run:\n\n{self.traceback}"
        return error_msg


def upload_pipeline(
    graph: Graph,
    name: str,
    environment_id_or_name: t.Union[str, int],
    required_gpu_vram_mb: t.Optional[int] = None,
    minimum_cache_number: t.Optional[int] = None,
    modules: t.Optional[t.List[str]] = None,
    accelerators: t.Optional[t.List[Accelerator]] | None = None,
    _metadata: t.Optional[dict] = None,
) -> pipeline_schemas.PipelineGet:
    if graph._has_run_startup:
        raise Exception("Graph has already been run, cannot upload")

    if modules is not None:
        for module_name in modules:
            module = importlib.import_module(module_name)
            cp.register_pickle_by_value(module)

    for variable in graph.variables:
        if isinstance(variable, Directory):
            if variable.remote_id is not None:
                continue
            variable_path = Path(variable.path)

            if not variable_path.exists():
                raise FileNotFoundError(
                    f"Directory not found for variable (path={variable.path}) "
                )
            if not variable_path.is_dir() and not str(variable_path).endswith(".zip"):
                raise ValueError(
                    f"Variable is not a directory or zip file (path={variable.path})"
                )

            zip_path = variable_path

            if not str(variable_path).endswith(".zip"):
                tmp_path = Path("/tmp") / (str(uuid.uuid4()) + ".zip")
                with ZipFile(str(tmp_path), "w") as zip_file:
                    for root, dirs, files in os.walk(str(variable_path)):
                        for file in files:
                            zip_file.write(
                                os.path.join(root, file),
                                arcname=file,
                            )
                zip_path = tmp_path

            zip_file = zip_path.open("rb")
            try:
                res = http.post_files(
                    "/v3/pipeline_files",
                    files=dict(pfile=zip_file),
                    progress=True,
                )

                new_path = res.json()["path"]
                variable.path = Path(new_path)
                variable.remote_id = res.json()["id"]
            except HTTPStatusError as e:
                if e.response.status_code == 403:
                    raise Exception(
                        f"Permission denied uploading file (path={variable.path}, "
                        f"variable={variable.title})"
                    )
            finally:
                zip_file.close()
        elif isinstance(variable, FileURL):
            try:
                urlparse(variable.url)
            except Exception:
                raise Exception(
                    f"Invalid URL for variable (url={variable.url})",
                )
        elif isinstance(variable, File):
            if variable.remote_id is not None:
                continue
            variable_path = Path(variable.path)

            if not variable_path.exists():
                raise FileNotFoundError(
                    f"File not found for variable (path={variable.path}, "
                    f"variable={variable.title})"
                )

            variable_file = variable_path.open("rb")
            try:
                res = http.post_files(
                    "/v3/pipeline_files",
                    files=dict(pfile=variable_file),
                    progress=True,
                )

                new_path = res.json()["path"]
                variable.path = Path(new_path)
                variable.remote_id = res.json()["id"]
            except HTTPStatusError as e:
                if e.response.status_code == 403:
                    raise Exception(
                        f"Permission denied uploading file (path={variable.path}, "
                        f"variable={variable.title})"
                    )
            finally:
                variable_file.close()

    params = dict()

    params["name"] = name
    params["environment_id_or_name"] = environment_id_or_name

    if required_gpu_vram_mb is not None:
        params["gpu_memory_min"] = required_gpu_vram_mb
    if minimum_cache_number is not None:
        params["minimum_cache_number"] = minimum_cache_number

    if accelerators is not None:
        params["accelerators"] = [
            acc.value for acc in accelerators if isinstance(acc, Accelerator)
        ]

    input_variables: t.List[dict] = []
    output_variables: t.List[dict] = []

    for variable in graph.variables:
        if variable.is_input:
            input_variables.append(variable.to_io_schema().json())

        if variable.is_output:
            output_variables.append(variable.to_io_schema().json())

    params["input_variables"] = input_variables
    params["output_variables"] = output_variables

    if _metadata is None:
        _metadata = dict()

    default_meta = dict(
        python_version=platform.python_version(),
        system=platform.system(),
        platform_version=platform.version(),
        platform=platform.platform(),
        sdk_version=version("pipeline-ai"),
    )

    _metadata.update(default_meta)

    params["_metadata"] = json.dumps(_metadata)

    graph_file = SpooledTemporaryFile()
    graph_file.write(cp.dumps(graph))
    graph_file.seek(0)
    try:
        res = http.post_files(
            "/v3/pipelines",
            files=dict(graph=graph_file),
            data=params,
        )
    except HTTPStatusError as e:
        if e.response.status_code == 403:
            raise Exception(
                e.response.text,
            )
        elif e.response.status_code == 400:
            raise Exception(
                e.response.text,
            )
        elif e.response.status_code == 409:
            raise Exception(
                e.response.text,
            )
        else:
            raise Exception(
                f"Unknown exception (code={e.response.status_code}), {e.response.text}"
            )

    if res.status_code != 200:
        print("Error uploading pipeline")
        print(res.text)
        res.raise_for_status()

    graph_file.close()

    raw_json = res.json()

    try:
        pointers_request = http.get(
            "/v3/pointers",
            params=dict(
                pipeline_name=name,
            ),
        )
        pointer_array = pointers_request.json()["data"]
        for item in pointer_array:
            if item["pipeline_id"] == raw_json["id"]:
                name = item["pointer"]

                break
    except HTTPStatusError:
        ...

    _print(f"Uploaded pipeline '{name}' with ID = {raw_json['id']}", level="SUCCESS")
    return pipeline_schemas.PipelineGet.parse_obj(raw_json)


def _data_to_run_input(data: t.Tuple) -> t.List[RunInput]:
    input_array = []

    for item in data:
        input_type = RunIOType.from_object(item)
        if input_type == RunIOType.file or isinstance(item, File):
            if isinstance(item, FileURL):
                input_schema = RunInput(
                    value=None,
                    type=RunIOType.file,
                    file_path=item.url,
                    file_name=item.url.split("/")[-1],
                )
                input_array.append(input_schema)
            else:
                file_obj = open(item.path, "rb") if isinstance(item, File) else item

                res = http.post_files(
                    "/v3/pipeline_files",
                    files=dict(pfile=file_obj),
                    progress=True,
                )
                file_path = res.json().get("path", None)

                if file_path is None:
                    raise Exception("Error uploading file")

                input_schema = RunInput(
                    value=None,
                    type=input_type,
                    file_path=file_path,
                    file_name=file_path.split("/")[-1],
                )
                input_array.append(input_schema)
            continue
        elif input_type == RunIOType.pkl:
            raise NotImplementedError("Python object input not yet supported")

        input_schema = RunInput(
            type=input_type,
            value=item,
            file_path=None,
            file_name=None,
        )
        input_array.append(input_schema)

    return input_array


def run_pipeline(
    pipeline_id_or_pointer: t.Union[str, int],
    *data,
    async_run: bool = False,
    return_response: bool = False,
) -> t.Union[Run, httpx.Response]:
    run_create_schema = RunCreate(
        pipeline_id_or_pointer=pipeline_id_or_pointer,
        input_data=_data_to_run_input(data),
        async_run=async_run if not current_configuration.is_debugging() else True,
    )

    res = http.post(
        "/v3/runs",
        json_data=run_create_schema.dict(),
        raise_for_status=False,
    )

    if return_response and not current_configuration.is_debugging():
        return res

    if res.status_code == 500:
        _print(
            f"Failed run (status={res.status_code}, text={res.text}, "
            f"headers={res.headers})",
            level="ERROR",
        )
        raise Exception(f"Error: {res.status_code}, {res.text}", res.status_code)
    elif res.status_code == 429:
        _print(
            f"Too many requests (status={res.status_code}, text={res.text})",
            level="ERROR",
        )
        raise Exception(
            "Too many requests, please try again later",
            res.status_code,
        )
    elif res.status_code == 404:
        _print(
            f"Pipeline not found (status={res.status_code}, text={res.text})",
            level="ERROR",
        )
        raise Exception("Pipeline not found", res.status_code)
    elif res.status_code == 503:
        _print(
            f"Environment not cached (status={res.status_code}, text={res.text})",
            level="ERROR",
        )
        raise Exception("Environment not cached", res.status_code)
    elif res.status_code == 502:
        _print(
            "Gateway error",
            level="ERROR",
        )
        raise Exception("Gateway error", res.status_code)

    run_get = Run.parse_obj(res.json())

    def _print_logs():
        for log in tail_run_logs(run_get.id):
            _print_remote_log(log)
            # _print(log)

    if current_configuration.is_debugging():
        log_thread = Thread(target=_print_logs)

        _print("Running in debug mode")
        _print(f"Run ID: {run_get.id}")

        def activate_log_thread(state: RunState):
            if state == RunState.running:
                log_thread.start()

        run_get = poll_async_run(
            run_get.id,
            interval=0.5,
            state_change_callback=activate_log_thread,
        )

        if run_get.state == RunState.completed:
            _print("Run completed successfully", level="SUCCESS")
            if run_get.result is not None:
                _print(f"Result array: {run_get.result.result_array()}")

    if RunState.is_terminal(run_get.state) and run_get.state != RunState.completed:
        error = getattr(run_get, "error", None)

        traceback = getattr(error, "traceback", None)
        exception = getattr(error, "exception", None)
        if (
            isinstance(exception, str)
            and "Tried to return a non-supported type from a pipeline" in exception
        ):
            raise Exception(
                f"Remote Run '{run_get.id}' failed because an unsupported type was returned. Check your outputs.\n(More information -> https://docs.mystic.ai/docs/inputoutpu-types)"  # noqa
            )

        raise Exception(
            f"Remote Run '{run_get.id}' failed with exception: "
            f"{traceback if traceback is not None and traceback != '' else exception}"
        )

    return run_get


def get_pipeline_run(run_id: str) -> Run:
    http_res = http.get(f"/v3/runs/{run_id}")

    run_get = Run.parse_obj(http_res.json())
    return run_get


def map_pipeline_mp(array: list, graph_id: str, *, pool_size=8):
    results = []

    num_batches = math.ceil(len(array) / pool_size)
    for b in range(num_batches):
        start_index = b * pool_size
        end_index = min(len(array), start_index + pool_size)

        inputs = [[graph_id, item] for item in array[start_index:end_index]]

        with Pool(pool_size) as p:
            batch_res = p.starmap(run_pipeline, inputs)
            results.extend(batch_res)

    return results


def stream_pipeline(
    pipeline_id_or_pointer: str,
    *data,
) -> t.Iterator[RunOutput]:
    run_create_schema = RunCreate(
        pipeline_id_or_pointer=pipeline_id_or_pointer,
        input_data=_data_to_run_input(data),
        async_run=False,
    )
    with http.stream_post(
        "/v3/runs/stream",
        json_data=run_create_schema.dict(),
    ) as generator:
        run_error = None
        for item in generator.iter_text():
            if item:
                try:
                    output = RunOutput.parse_raw(item)
                    yield output
                except ValidationError:
                    run_error = RunError.parse_raw(item)
                if run_error is not None:
                    raise PipelineRunError(
                        run_error.exception, traceback=run_error.traceback
                    )


def poll_async_run(
    run_id: str,
    *,
    timeout: int | None = None,
    interval: float | int = 1.0,
    state_change_callback: t.Callable[[RunState], None] | None = None,
) -> Run:
    start_time = time.time()
    current_state = None
    while True:
        run_get = get_pipeline_run(run_id)
        if current_state != run_get.state and current_configuration.is_debugging():
            _print(f"Run state: {run_get.state.name}")
            current_state = run_get.state
            if state_change_callback is not None:
                state_change_callback(run_get.state)
        if run_get.state in [
            RunState.completed,
            RunState.failed,
            RunState.rate_limited,
            RunState.lost,
            RunState.no_environment_installed,
        ]:
            return run_get

        if timeout is not None and time.time() - start_time > timeout:
            raise TimeoutError(f"Timeout waiting for run (run_id={run_id})")

        time.sleep(interval)



import io
import json
import typing as t
from datetime import datetime
from enum import Enum

from pydantic import BaseModel


class RunState(int, Enum):
    created: int = 0
    routing: int = 1
    resource_accepted: int = 2
    # this includes creating a virtual environment and installing
    # packages
    creating_environment: int = 3
    # starting subrprocess running worker in custom environment
    starting_worker: int = 4
    downloading_graph: int = 5
    caching_graph: int = 6
    running: int = 7
    resource_returning: int = 8
    api_received: int = 9
    celery_worker_received: int = 22

    in_queue: int = 16
    denied: int = 11
    resource_rejected: int = 14
    resource_died: int = 15
    retrying: int = 13
    rerouting: int = 21

    completed: int = 10
    failed: int = 12
    rate_limited: int = 17
    lost: int = 18
    no_environment_installed: int = 19

    unknown: int = 20

    @staticmethod
    def is_terminal(state: "RunState") -> bool:
        return state in [
            RunState.completed,
            RunState.failed,
            RunState.lost,
            RunState.no_environment_installed,
            RunState.rate_limited,
        ]

    @classmethod
    def __get_validators__(cls):
        cls.lookup = {v: k.value for v, k in cls.__members__.items()}
        cls.value_lookup = {k.value: v for v, k in cls.__members__.items()}
        yield cls.validate

    @classmethod
    def validate(cls, v):
        try:
            v = int(v)
        except Exception:
            ...

        if isinstance(v, str):
            return cls.lookup[v]
        elif isinstance(v, int):
            return getattr(cls, cls.value_lookup[v])
        else:
            raise ValueError(f"Invalid value: {v}")


class RunErrorType(Enum):
    input_error = 1
    unroutable = 2
    graph_error = 3
    runtime_error = 4


class RunError(BaseModel):
    exception: str
    traceback: t.Optional[str]


class RunFileType(Enum):
    input = "input"
    output = "output"


class RunFile(BaseModel):
    id: str
    run_id: str
    io_type: RunFileType
    path: str

    class Config:
        # use_enum_values = True
        orm_mode = True


class RunIOType(str, Enum):
    integer: str = "integer"
    string: str = "string"
    fp: str = "fp"
    dictionary: str = "dictionary"
    boolean: str = "boolean"
    none: str = "none"
    array: str = "array"

    pkl: str = "pkl"
    file: str = "file"

    @classmethod
    def from_object(cls, obj: t.Any):
        # Get the enum type for the object.
        from pipeline.objects.graph import File

        if isinstance(obj, int) or obj is int:
            return cls.integer
        elif isinstance(obj, float) or obj is float:
            return cls.fp
        elif isinstance(obj, str) or obj is str:
            return cls.string
        elif isinstance(obj, bool) or obj is bool:
            return cls.boolean
        elif obj is None:
            return cls.none
        elif isinstance(obj, dict) or obj is dict:
            if obj is dict:
                return cls.dictionary
            try:
                json.dumps(obj)
            except (TypeError, OverflowError):
                return cls.pkl
            return cls.dictionary
        elif isinstance(obj, list) or obj is list:
            if obj is list:
                return cls.array
            try:
                json.dumps(obj)
            except (TypeError, OverflowError):
                return cls.pkl
            return cls.array
        elif isinstance(obj, io.BufferedIOBase) or obj is File:
            return cls.file
        else:
            return cls.pkl

    @staticmethod
    def to_object(io_type: "RunIOType"):
        if isinstance(io_type, str):
            io_type = RunIOType(io_type)

        if io_type == RunIOType.integer:
            return int
        elif io_type == RunIOType.fp:
            return float
        elif io_type == RunIOType.string:
            return str
        elif io_type == RunIOType.boolean:
            return bool
        elif io_type == RunIOType.none:
            return None
        elif io_type == RunIOType.dictionary:
            return dict
        elif io_type == RunIOType.array:
            return list
        else:
            raise ValueError(f"Invalid io_type: {io_type}")


class RunOutputFile(BaseModel):
    name: str
    path: str
    url: str
    size: int


class RunOutput(BaseModel):
    type: RunIOType
    value: t.Optional[t.Any]
    file: t.Optional[RunOutputFile]


class RunResult(BaseModel):
    run_id: str
    outputs: t.List[RunOutput]

    def result_array(self) -> t.List[t.Any]:
        return [output.value for output in self.outputs]


class Run(BaseModel):
    id: str

    created_at: datetime

    pipeline_id: str
    environment_id: str
    environment_hash: str

    state: RunState

    error: t.Optional[RunError]

    result: t.Optional[RunResult]

    class Config:
        # use_enum_values = True
        orm_mode = True


class RunStateTransition(BaseModel):
    run_id: str
    new_state: RunState
    time: datetime


class RunInput(BaseModel):
    type: RunIOType
    value: t.Any

    file_name: t.Optional[str]
    file_path: t.Optional[str]


class RunCreate(BaseModel):
    pipeline_id_or_pointer: str
    input_data: t.List[RunInput]
    async_run: bool = False



from datetime import datetime

from pipeline.cloud.schemas import BaseModel


class EnvironmentGet(BaseModel):
    id: str
    created_at: datetime
    updated_at: datetime

    name: str
    python_requirements: list[str]
    hash: str


class EnvironmentCreate(BaseModel):
    name: str
    python_requirements: list[str]



import importlib
import importlib.metadata
import io
import random
import string
from typing import Any, Callable, List, Optional, Union

from cloudpickle import dumps, register_pickle_by_value
from dill import loads


def package_version() -> str:
    """Return the version of the installed `pipeline-ai` package."""
    return importlib.metadata.version("pipeline-ai")


def generate_id(length: int) -> str:
    return "".join((random.choice(string.ascii_letters) for i in range(length)))


def python_object_to_hex(
    obj: Any,
    modules: Optional[List[str]] = None,
) -> str:
    return dump_object(obj, modules=modules).hex()


def hex_to_python_object(hex: str) -> Any:
    h = bytes.fromhex(hex)
    return loads(h)


def dump_object(obj: Any, modules: Optional[List[str]] = None) -> bytes:
    """Serialize `obj` as bytes."""
    if modules is not None:
        for module_name in modules:
            module = importlib.import_module(module_name)
            register_pickle_by_value(module)
    return dumps(obj)


def load_object(pickled: Union[bytes, str]) -> Any:
    """Deserialize an object from the payload."""
    if isinstance(pickled, str):
        return hex_to_python_object(pickled)
    else:
        return loads(pickled)


def python_object_to_name(obj: Any) -> Optional[str]:
    # Consider limiting the size of the name in future releases
    name = getattr(obj, "__name__", str(obj))

    return name


# def python_object_to_file_create(obj: Any, name: str = None):
#     if name is None:
#         name = generate_id(20)

#     return FileCreate(name=name, file_bytes=python_object_to_hex)


class CallbackBytesIO(io.BytesIO):
    """Provides same interface as BytesIO but additionally calls a callback function
    whenever the 'read' method is called.

    This is similar to tqdm's own CallbackIOWrapper but this does not play nicely with
    all features of httpx so we use our own in some cases.
    """

    def __init__(self, callback: Callable, initial_bytes: bytes):
        self._callback = callback
        super().__init__(initial_bytes)

    def read(self, size=-1) -> bytes:
        data = super().read(size)
        self._callback(len(data))
        return data



from pipeline.configuration import current_configuration
from pipeline.objects import (
    File,
    FileURL,
    Pipeline,
    Variable,
    entity,
    onnx_to_pipeline,
    pipe,
)

__all__ = [
    "Pipeline",
    "Variable",
    "entity",
    "pipe",
    "File",
    "FileURL",
    "onnx_to_pipeline",
    "current_configuration",
]



import typing as t

import httpx

from pipeline.cloud import http
from pipeline.cloud.pipelines import _data_to_run_input
from pipeline.cloud.schemas.runs import Run, RunCreate, RunState
from pipeline.util.logging import _print


async def run_pipeline(
    pipeline_id_or_pointer: t.Union[str, int],
    *data,
    async_run: bool = False,
    return_response: bool = False,
) -> t.Union[Run, httpx.Response]:
    run_create_schema = RunCreate(
        pipeline_id_or_pointer=pipeline_id_or_pointer,
        input_data=_data_to_run_input(data),
        async_run=async_run,
    )

    res = await http.async_post(
        "/v3/runs",
        json_data=run_create_schema.dict(),
        raise_for_status=False,
    )

    if return_response:
        return res

    if res.status_code == 500:
        _print(
            f"Failed run (status={res.status_code}, text={res.text}, "
            f"headers={res.headers})",
            level="ERROR",
        )
        raise Exception(f"Error: {res.status_code}, {res.text}", res.status_code)
    elif res.status_code == 429:
        _print(
            f"Too many requests (status={res.status_code}, text={res.text})",
            level="ERROR",
        )
        raise Exception(
            "Too many requests, please try again later",
            res.status_code,
        )
    elif res.status_code == 404:
        _print(
            f"Pipeline not found (status={res.status_code}, text={res.text})",
            level="ERROR",
        )
        raise Exception("Pipeline not found", res.status_code)
    elif res.status_code == 503:
        _print(
            f"Environment not cached (status={res.status_code}, text={res.text})",
            level="ERROR",
        )
        raise Exception("Environment not cached", res.status_code)
    elif res.status_code == 502:
        _print(
            "Gateway error",
            level="ERROR",
        )
        raise Exception("Gateway error", res.status_code)

    run_get = Run.parse_obj(res.json())

    if RunState.is_terminal(run_get.state) and run_get.state != RunState.completed:
        raise Exception(
            f"Remote Run '{run_get.id}' failed with exception: "
            f"{getattr(getattr(run_get, 'error', None), 'traceback', None)}"
        )

    return run_get



import io
import os
import typing as t

import httpx

# from httpx import _GeneratorContextManager
import requests
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor
from tqdm import tqdm

from pipeline import current_configuration
from pipeline.util.logging import PIPELINE_STR

_client = None
_client_async = None


def _get_client() -> httpx.Client:
    global _client
    if _client is None:
        url = (
            active_remote.url
            if (active_remote := current_configuration.active_remote) is not None
            else os.environ.get("PIPELINE_API_URL", "https://www.mystic.ai/")
        )
        token = (
            current_configuration.active_remote.token
            if current_configuration.active_remote is not None
            else os.environ.get("PIPELINE_API_TOKEN", None)
        )
        _client = httpx.Client(
            base_url=url,
            headers={
                "Authorization": f"Bearer {token}",
            },
            timeout=300,
        )
    return _client


def _get_async_client() -> httpx.AsyncClient:
    global _client_async
    if _client_async is None:
        url = (
            active_remote.url
            if (active_remote := current_configuration.active_remote) is not None
            else os.environ.get("PIPELINE_API_URL", "https://www.mystic.ai/")
        )
        token = (
            current_configuration.active_remote.token
            if current_configuration.active_remote is not None
            else os.environ.get("PIPELINE_API_TOKEN", None)
        )
        _client_async = httpx.AsyncClient(
            base_url=url,
            headers={
                "Authorization": f"Bearer {token}",
            },
            timeout=300,
        )
    return _client_async


def post(
    endpoint: str,
    json_data: dict = None,
    raise_for_status: bool = True,
) -> httpx.Response:
    client = _get_client()
    response = client.post(
        endpoint,
        json=json_data,
    )
    if raise_for_status:
        response.raise_for_status()

    return response


async def async_post(
    endpoint: str,
    json_data: dict = None,
    raise_for_status: bool = True,
) -> httpx.Response:
    client = _get_async_client()
    response = await client.post(
        endpoint,
        json=json_data,
    )
    if raise_for_status:
        response.raise_for_status()

    return response


def patch(
    endpoint: str,
    json_data: dict = None,
    raise_for_status: bool = True,
) -> httpx.Response:
    client = _get_client()
    response = client.patch(
        endpoint,
        json=json_data,
    )
    if raise_for_status:
        response.raise_for_status()

    return response


def get(
    endpoint: str,
    **kwargs,
) -> httpx.Response:
    client = _get_client()
    response = client.get(endpoint, **kwargs)
    response.raise_for_status()

    return response


def delete(
    endpoint: str,
    **kwargs,
) -> httpx.Response:
    client = _get_client()
    response = client.delete(endpoint, **kwargs)
    response.raise_for_status()

    return response


def create_callback(encoder: MultipartEncoder) -> t.Callable:
    encoder_len = encoder.len
    bar = tqdm(
        desc=f"{PIPELINE_STR} Uploading",
        unit="B",
        unit_scale=True,
        total=encoder_len,
        unit_divisor=1024,
    )

    def callback(monitor):
        bar.n = monitor.bytes_read
        bar.refresh()
        if monitor.bytes_read == encoder_len:
            bar.close()

    return callback


def get_progress_bar_uploader(
    files: t.Dict[str, io.BufferedIOBase], params: t.Dict[str, str]
) -> MultipartEncoderMonitor:
    encoder = create_upload(files, params)
    callback = create_callback(encoder)
    monitor = MultipartEncoderMonitor(encoder, callback)
    return monitor


def create_upload(
    files: t.Dict[str, io.BufferedIOBase], params: t.Dict[str, str]
) -> MultipartEncoder:
    file_dict = {
        param_key: (
            getattr(file, "name", "undefined"),
            file,
            "application/octet-stream",
            {
                "Content-Transfer-Encoding": "binary",
            },
        )
        for param_key, file in files.items()
    }

    return MultipartEncoder(
        dict(
            **file_dict,
            **params,
        ),
    )


def post_files(
    endpoint: str,
    files: t.Dict[str, t.Any],
    params: t.Dict[str, t.Any] = dict(),
    data: t.Dict[str, t.Any] = dict(),
    progress: bool = False,
) -> httpx.Response:
    if progress:
        monitor = get_progress_bar_uploader(files=files, params=params)
        url = (
            active_remote.url
            if (active_remote := current_configuration.active_remote) is not None
            else os.environ.get("PIPELINE_API_URL", "https://www.mystic.ai/")
        )
        token = (
            current_configuration.active_remote.token
            if current_configuration.active_remote is not None
            else os.environ.get("PIPELINE_API_TOKEN", None)
        )
        response = requests.post(
            f"{url}{endpoint}",
            data=monitor,
            headers={
                "content-Type": monitor.content_type,
                "Authorization": f"Bearer {token}",
            },
            params=params,
        )
    else:
        client = _get_client()
        response = client.post(
            endpoint,
            files=files,
            data=data,
            params=params,
        )

    return response


def stream_post(
    endpoint: str,
    json_data: dict = None,
) -> t.Iterator[httpx.Response]:
    client = _get_client()
    return client.stream(
        "POST",
        endpoint,
        json=json_data,
    )



from httpx import HTTPStatusError

from pipeline.cloud import http
from pipeline.util.logging import _print


def create_environment(
    name: str,
    python_requirements: list[str],
    allow_existing: bool = True,
) -> int:
    try:
        res = http.post(
            "/v3/environments",
            json_data={"name": name, "python_requirements": python_requirements},
        )

        env_id = res.json()["id"]
        _print(f"Created environment '{name}' with ID = {env_id}", level="SUCCESS")
    except HTTPStatusError as e:
        if e.response.status_code == 409 and allow_existing:
            try:
                res = http.get(
                    f"/v3/environments/{name}",
                )
                env_id = res.json()["id"]
                _print(
                    f"Using existing environment {name} with ID = {env_id}",
                    level="INFO",
                )
            except HTTPStatusError as e2:
                if e2.response.status_code == 404:
                    raise Exception(f"Environment {name} does not exist")
                elif e2.response.status_code == 401:
                    # raise Exception(f"Environment {name} is not accessible to you.")
                    _print(
                        "Environment already exists and is not accessible to you.",
                        level="WARNING",
                    )
                    env_id = name
                else:
                    raise Exception(
                        f"Unknown error: {e2.response.status_code}, {e2.response.text}"
                    )
        else:
            raise e

    return env_id
